import os.path
import re
from copy import deepcopy
from typing import Any, Dict

import toml
from scalpl import Cut

from stela.utils import get_base_path, merge_dicts


def read_embed(options: "StelaOptions") -> Dict[Any, Any]:
    """Stela Embed Loader.

    Will be removed on 6.0

    :param options: StelaOptions instance
    :return: Dict
    """
    from loguru import logger

    # Get toml data
    path = get_base_path()
    filepath = path.joinpath("pyproject.toml")
    if not os.path.exists(filepath):
        return {}
    toml_data = toml.load(filepath)
    table = options.env_table

    # Get environment layers
    with open(filepath, "r") as file:
        raw_data = file.read()
    pattern = rf"(?<=\[{table}\.)\w+"
    environment_list = re.findall(pattern, raw_data)

    logger.info(f"Looking for table [{table}] inside pyproject.toml...")
    proxy = Cut(toml_data)
    table_data = proxy.get(options.env_table, {})
    first_level_data = deepcopy(table_data)

    #  remove keys which represent environments in first level
    keys = list(first_level_data.keys())
    for key in keys:
        if key in environment_list:
            del first_level_data[key]

    if options.use_environment_layers:
        sub_table_name = f"{table}.{options.current_environment}"
        logger.info(
            f"Looking for sub-table [{sub_table_name}] inside pyproject.toml..."
        )
        sub_table = proxy.get(sub_table_name, {})
        merge_dicts(source_dict=sub_table, target_dict=first_level_data)
    return first_level_data



import os
from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from loguru import logger

from stela.exceptions import StelaEnvironmentNotFoundError, StelaFileTypeError
from stela.parsers.dotenv import read_dotenv
from stela.parsers.other_files import StelaFileReader
from stela.utils import StelaFileType

DEFAULT_ORDER = ["embed", "file", "custom"]


@dataclass
class StelaBaseOptions:
    config_file_extension: StelaFileType = StelaFileType.INI
    env_file: str = ".env"
    dotenv_overwrites_memory: bool = False
    log_filtered_value: bool = True
    show_logs: bool = False
    config_file_path: str = "."
    evaluate_data: bool = False
    default_environment: Optional[str] = None
    warn_if_env_is_missing: bool = False
    dotenv_encoding: str = "utf-8"
    _dotenv_data: Optional[Dict[Any, Any]] = None
    current_environment: Optional[str] = None
    env_table: str = "environment"
    _filenames: List[str] = field(default_factory=list)

    @classmethod
    def get_from_env_or_settings(
        cls, key: str, file_settings: Dict[str, Any], default: Any
    ) -> Any:
        """Get from environment or settings the Stela Option argument.

        :param key: Current stela argument.
        :param file_settings: Stela Data from pyproject.toml.
        :param default: default value for argument.
        :return: Any
        """
        environment_name = f"STELA_{key.upper().replace('.', '_')}"
        value = os.getenv(environment_name, file_settings.get(key, default))
        if value and isinstance(value, str) and isinstance(default, bool):
            return value.lower() == "true"
        return value

    @classmethod
    def _get_current_environment(cls, settings):
        # Get from Memory
        settings["current_environment"] = os.getenv(
            settings["environment_variable_name"]
        )

        # Get from .env if available
        if not settings["current_environment"]:
            settings["current_environment"] = settings["_dotenv_data"].get(
                settings["environment_variable_name"]
            )

        # Get from Default
        if not settings["current_environment"]:
            settings["current_environment"] = settings["default_environment"]

        # No environment found.
        if not settings["current_environment"] and settings.get(
            "use_environment_layers"
        ):
            raise StelaEnvironmentNotFoundError("Environment not found.")

    @classmethod
    def _get_dotenv_data(cls, settings, update_environs: bool):
        """Get data from dotenv files."""
        if settings.get("do_not_read_dotenv"):
            settings["_dotenv_data"] = {}
        else:
            settings["_dotenv_data"] = read_dotenv(
                config_file_path=settings["config_file_path"],
                env_file=settings["env_file"],
                overwrites_memory=settings["dotenv_overwrites_memory"],
                encoding=settings["dotenv_encoding"],
                verbose=settings["warn_if_env_is_missing"],
                update_environs=update_environs,
                filter_logs=settings["log_filtered_value"],
            )

    def get_extensions(self) -> List[str]:
        """Return file extensions for project configuration files."""
        return self.config_file_extension.value  # type: ignore

    @classmethod
    def get_config(cls) -> "StelaBaseOptions":
        """Get config from pyproject.toml."""

        file_settings, settings = cls.get_settings()
        try:
            config_file_extension = cls.get_from_env_or_settings(
                "config_file_extension",
                file_settings,
                cls.config_file_extension.value[0].replace(".", "").upper(),
            )
            settings["config_file_extension"] = StelaFileType[config_file_extension]
        except KeyError:
            raise StelaFileTypeError(
                f"Invalid file type: {file_settings.get('config_file_extension')}"
            )
        cls._get_dotenv_data(settings, update_environs=False)
        cls._get_current_environment(settings)
        settings["_filenames"] = [
            f"{settings.get('config_file_prefix','')}{settings['current_environment']}{extension}"
            for extension in settings["config_file_extension"].value
        ]
        return cls(**settings)

    @classmethod
    @abstractmethod
    def get_settings(cls):
        raise NotImplementedError()


@dataclass
class StelaOptions(StelaBaseOptions):
    """Stela Options data class for Dot class."""

    environment_variable_name: str = "STELA_ENV"
    final_loader: str = "stela.main.loader.default_loader"
    raise_on_missing_variable: bool = True
    evaluate_data: bool = True
    dotenv_overwrites_memory: bool = True
    default_environment: Optional[str] = None

    @classmethod
    def get_settings(cls):
        file_settings = {}
        toml_path = "pyproject.toml"
        reader = StelaFileReader()
        if os.path.exists(toml_path):
            toml_settings = reader.load_toml(toml_path)
            file_settings = toml_settings.get("tool", {}).get("stela", {})
            if file_settings:
                logger.info("Using pyproject.toml for stela settings.")
        else:
            ini_path = ".stela"
            if os.path.exists(ini_path):
                ini_settings = reader.load_ini(ini_path)
                file_settings = ini_settings.get("stela", {})
                if file_settings:
                    logger.info("Using .stela for stela settings.")
        if not file_settings:
            logger.info(
                "No stela file configuration found. Using default stela settings."
            )
        settings = {
            "environment_variable_name": cls.get_from_env_or_settings(
                "environment_variable_name",
                file_settings,
                cls.environment_variable_name,
            ),
            "default_environment": cls.get_from_env_or_settings(
                "default_environment", file_settings, cls.default_environment
            ),
            "evaluate_data": cls.get_from_env_or_settings(
                "evaluate_data", file_settings, cls.evaluate_data
            ),
            "config_file_path": cls.get_from_env_or_settings(
                "config_file_path", file_settings, cls.config_file_path
            ),
            "show_logs": cls.get_from_env_or_settings(
                "show_logs", file_settings, cls.show_logs
            ),
            "log_filtered_value": cls.get_from_env_or_settings(
                "log_filtered_value",
                file_settings,
                cls.log_filtered_value,
            ),
            "env_file": cls.get_from_env_or_settings(
                "env_file", file_settings, cls.env_file
            ),
            "warn_if_env_is_missing": cls.get_from_env_or_settings(
                "warn_if_env_is_missing", file_settings, cls.warn_if_env_is_missing
            ),
            "dotenv_encoding": cls.get_from_env_or_settings(
                "dotenv_encoding", file_settings, cls.dotenv_encoding
            ),
            "final_loader": cls.get_from_env_or_settings(
                "final_loader", file_settings, cls.final_loader
            ),
            "dotenv_overwrites_memory": cls.get_from_env_or_settings(
                "dotenv_overwrites_memory", file_settings, cls.dotenv_overwrites_memory
            ),
            "raise_on_missing_variable": cls.get_from_env_or_settings(
                "raise_on_missing_variable",
                file_settings,
                cls.raise_on_missing_variable,
            ),
        }
        return file_settings, settings



"""Import Stela Data.

If you need to reload data inside you application,
use the stela_reload function.

"""
import os
from ast import literal_eval
from typing import Any, List

from loguru import logger

from stela.config import StelaOptions
from stela.config.cut import StelaCutOptions
from stela.config.stub import create_stela_stub
from stela.exceptions import StelaValueError
from stela.main.cut import StelaCut, StelaCutMain
from stela.main.dot import StelaDotMain
from stela.utils import find_file_folder, show_value

__version__ = "5.1.4"


def _get_stela_cut() -> StelaCutMain:
    stela_config = StelaCutOptions.get_config()

    conf_path = find_file_folder("conf_stela.py")
    if conf_path:
        try:
            import conf_stela  # noqa
        except ImportError:
            pass

    if stela_config.show_logs:
        logger.enable("stela")
    else:
        logger.disable("stela")

    stela = StelaCutMain(options=stela_config)

    return stela


def _get_stela() -> "Stela":
    stela_config = StelaOptions.get_config()

    if stela_config.show_logs:
        logger.enable("stela")
    else:
        logger.disable("stela")

    stela_data = StelaDotMain(options=stela_config)
    stela_data.get_project_settings()

    class Stela:
        __slots__ = (
            [str(k) for k in stela_data.settings.keys()]
            if stela_data.settings.keys()
            else ["_NO_ENV_FOUND_"]
        )
        _stela_options: StelaOptions = stela_config
        _stela_data: StelaDotMain = stela_data

        def _get_attributes(self, current_obj: object, data_dict: dict):
            for attr, value in data_dict.items():
                if isinstance(value, dict):

                    class StelaNestedObj:
                        __slots__ = [str(k) for k in value.keys()]

                    nested_obj = StelaNestedObj()
                    self._get_attributes(current_obj=nested_obj, data_dict=value)
                    setattr(current_obj, attr, nested_obj)
                else:
                    if stela_config.dotenv_overwrites_memory:
                        current_value = value
                    else:
                        current_value = os.getenv(attr, value)
                    if stela_config.evaluate_data:
                        try:
                            current_value = literal_eval(current_value)
                        except Exception:
                            logger.debug(f"Can't eval value for: {attr}")
                    setattr(current_obj, attr, current_value)

        @property
        def current_environment(self):
            return self._stela_options.current_environment or "GLOBAL"

        @property
        def default_environment(self):
            return self._stela_options._default_environment or "GLOBAL"

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

            self._get_attributes(current_obj=self, data_dict=stela_data.settings)

        def __getattribute__(self, item):
            # return from os.environ if exists
            if "item" in os.environ:
                value = os.environ[item]
                logger.debug(
                    f"Using environment value: {item}={show_value(value, stela_config.log_filtered_value)}"
                )
                return value
            try:
                value = super().__getattribute__(item)
                if not item.startswith("_"):
                    logger.debug(
                        f"Using stela value: {item}={show_value(value, stela_config.log_filtered_value)}"
                    )
                return value
            except AttributeError as exc:
                if stela_config.raise_on_missing_variable:
                    raise StelaValueError(
                        f"Stela did not found value for {item}."
                    ) from exc
                logger.warning(f"Stela did not found value for {item}. Returning None.")
                return None

        def __setattr__(self, key, value):
            if key in self.__slots__ and hasattr(self, key):
                raise StelaValueError(f"Attribute {key} is ready-only.")
            super().__setattr__(key, value)

        def get(self, var_name: str, raise_on_missing: bool = True):
            if raise_on_missing:
                try:
                    return getattr(self, var_name)
                except AttributeError as exc:
                    raise StelaValueError(
                        f"Stela did not found value for {var_name}."
                    ) from exc
            return getattr(self, var_name, None)

        def get_or_default(self, var_name: str, default: Any):
            return getattr(self, var_name, default)

        def list(self) -> List[str]:
            return [k for k in self._stela_data.settings.keys()]

    create_stela_stub(stela_data.settings)

    return Stela()


settings: StelaCut = _get_stela_cut().get_project_settings()  # Will be removed on 6.0

env = _get_stela()



from typing import Any, Dict

from stela.config import StelaOptions


def default_loader(options: StelaOptions, env_data: Dict[str, Any]) -> Dict[str, Any]:
    """Stela Default Loader.

    :param options: Stela Options
    :param env_data: Dict with environment data
    :return: Dict
    """
    from loguru import logger

    logger.info(
        f"Using Stela Default Loader. Current environment is: {options.current_environment}"
    )
    return env_data



from typing import Any, Callable, Dict, Optional
from warnings import warn

from stela.exceptions import StelaTooManyLoadersError
from stela.loaders import SingletonMeta

WARN_MESSAGE = "Add Loader is deprecated and will be removed in 6.0. Please use the conf_stela.py file instead."


class StelaCutLoader(metaclass=SingletonMeta):
    """Stela Loader for full lifecycle.

    Will be removed on 6.0

    """

    def __init__(self):
        self.pre_load_function: Optional[Callable] = None
        self.custom_load_function: Optional[Callable] = None
        self.post_load_function: Optional[Callable] = None

        self.pre_data: Optional[Dict[Any, Any]] = None
        self.embed_data: Optional[Dict[Any, Any]] = None
        self.file_data: Optional[Dict[Any, Any]] = None
        self.custom_data: Optional[Dict[Any, Any]] = None
        self.post_data: Optional[Dict[Any, Any]] = None

    def add_pre_loader(self, function: Callable):
        warn(
            WARN_MESSAGE,
            DeprecationWarning,
            stacklevel=2,
        )
        if not self.pre_load_function:
            self.pre_load_function = function
        else:
            raise StelaTooManyLoadersError(
                f"A previous pre-load function was added: {self.pre_load_function.__name__}."
            )

    def add_custom_loader(self, function: Callable):
        warn(
            WARN_MESSAGE,
            DeprecationWarning,
            stacklevel=2,
        )
        if not self.custom_load_function:
            self.custom_load_function = function
        else:
            raise StelaTooManyLoadersError(
                f"A previous custom-load was added {self.custom_load_function.__name__}."
            )

    def add_post_loader(self, function: Callable):
        warn(
            WARN_MESSAGE,
            DeprecationWarning,
            stacklevel=2,
        )
        if not self.post_load_function:
            self.post_load_function = function
        else:
            raise StelaTooManyLoadersError(
                f"A previous post-load was added {self.post_load_function.__name__}."
            )



"""Stela Decorators."""
from functools import wraps


def pre_load(f):
    """Pre-Loader decorator.

    Use this decorator to handle data
    before invoking
    the custom load or default_loader.

    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.loaders.cut import StelaCutLoader

    loader = StelaCutLoader()
    loader.add_pre_loader(f)

    return wrapper


def custom_load(f):
    """Run Custom Loader decorator.

    This loader will be invoked instead
    the default_loader.

    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.loaders.cut import StelaCutLoader

    loader = StelaCutLoader()
    loader.add_custom_loader(f)

    return wrapper


def post_load(f):
    """Post-Loader decorator.

    Use this decorator to handle data
    after invoking
    the custom load or default_loader.

    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.loaders.cut import StelaCutLoader

    loader = StelaCutLoader()
    loader.add_post_loader(f)

    return wrapper


def stela_enable_logs(f):
    from loguru import logger

    from stela import env

    @wraps(f)
    def wrapper(*args, **kwargs):
        logger.enable("stela")
        ret = f(*args, **kwargs)
        logger.disable("stela")
        if env._stela_options.show_logs:
            logger.enable("stela")
        return ret

    return wrapper


def stela_disable_logs(f):
    from loguru import logger

    from stela import env

    @wraps(f)
    def wrapper(*args, **kwargs):
        logger.disable("stela")
        ret = f(*args, **kwargs)
        if env._stela_options.show_logs:
            logger.enable("stela")
        return ret

    return wrapper



import configparser
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Tuple

import toml
import yaml

from stela.utils import get_base_path


@dataclass
class StelaFileReader:
    """Read other files to load settings.

    Will be removed on 6.0
    """

    def load_from_file(self, filepath: Path, options: "StelaOptions") -> Dict[Any, Any]:
        """Resolve correct function for file extension."""
        function_name = (
            f"load_{options.config_file_extension.value[0].replace('.', '')}"
        )
        return getattr(self, function_name)(filepath)  # type: ignore

    def load_ini(self, filepath: Path) -> Dict[Any, Any]:
        """Load INI files.

        :param filepath: Path instance
        :return: Dict
        """
        config = configparser.ConfigParser()
        config.read(filepath)
        ini_settings: Dict[Any, Any] = {}
        for main_key in config.keys():
            ini_settings[main_key] = {}
            for key in config[main_key].keys():
                ini_settings[main_key][key] = config[main_key][key]

        return ini_settings

    def load_yaml(self, filepath: Path) -> Dict[Any, Any]:
        """Load YAML files.

        :param filepath: Path instance
        :return: Dict
        """
        with open(filepath, "r") as yaml_file:
            config = yaml.safe_load(yaml_file)
            return config  # type: ignore

    def load_toml(self, filepath: Path) -> Dict[Any, Any]:
        """Load TOML files.

        :param filepath: Path instance
        :return: Dict
        """
        config = toml.load(filepath)
        return config  # type: ignore

    def load_json(self, filepath: Path) -> Dict[Any, Any]:
        """Load JSON files.

        :param filepath: Path instance
        :return: Dict
        """
        with open(filepath, "r") as json_file:
            config = json.load(json_file)
            return config  # type: ignore


def read_file(options: "StelaCutOptions") -> Tuple[str, Dict[Any, Any]]:
    """Stela File Loader.

    Will be removed on 6.0

    :param options: StelaOptions instance
    :return: Dict
    """
    from loguru import logger

    path = get_base_path()
    reader = StelaFileReader()
    settings_data = {}
    file_name = ""
    for filename in options._filenames:
        filepath = path.joinpath(options.config_file_path, filename)
        if filepath.exists():
            logger.info(f"Reading file {filepath}...")
            settings_data = reader.load_from_file(filepath, options)
            file_name = "multiple-files" if len(options._filenames) > 1 else filename
    return file_name, settings_data



from abc import abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

from loguru import logger

from stela.config import StelaOptions
from stela.config.cut import StelaCutOptions
from stela.exceptions import StelaEnvironmentNotFoundError
from stela.loaders.cut import StelaCutLoader
from stela.utils import show_value


@dataclass
class StelaBaseMain:
    options: Union[StelaCutOptions, StelaOptions]
    loader: Optional[StelaCutLoader] = None
    settings: Dict[Any, Any] = field(default_factory=dict)

    def log_current_data(self, origin: str) -> None:
        def log_dict(dict_key: Any, dict_value: Any, parents: List):
            if isinstance(dict_value, dict):
                for k, v in dict_value.items():
                    if dict_key not in parents:
                        parents.append(dict_key)
                    log_dict(k, v, parents)
            else:
                parents_key = (
                    parents + [dict_key] if dict_key not in parents else parents
                )
                logger.debug(
                    f"[{origin}] {'.'.join(parents_key)} = "
                    f"{show_value(dict_value, self.options.log_filtered_value)}"
                )

        for key, value in self.settings.items():
            log_dict(key, value, [])

    @property
    def environment(self) -> str:
        """Return Current Environment."""
        if not self.options.current_environment:
            raise StelaEnvironmentNotFoundError("Environment not found.")
        return self.options.current_environment

    @abstractmethod
    def get_project_settings(self):
        """Get project settings running Stela Lifecycle.

        :return: Dict
        """



"""Stela Options module."""
from dataclasses import dataclass, field
from typing import List

import toml

from stela.config import DEFAULT_ORDER, StelaBaseOptions
from stela.utils import find_file_folder


@dataclass
class StelaCutOptions(StelaBaseOptions):
    """Stela Options data class for Cut class.

    Will be removed on 6.0
    """

    environment_variable_name: str = "ENVIRONMENT"

    do_not_read_dotenv: bool = False
    load_order: List[str] = field(default_factory=list)
    use_environment_layers: bool = False
    do_not_read_environment: bool = False
    environment_suffix: str = ""
    environment_prefix: str = ""
    config_file_prefix: str = ""
    config_file_suffix: str = ""

    @classmethod
    def get_settings(cls):
        file_settings = {}
        path = find_file_folder("pyproject.toml")
        if path:
            filepath = path.joinpath("pyproject.toml")
            toml_settings = toml.load(filepath)
            file_settings = toml_settings.get("tool", {}).get("stela", {})
        settings = {
            "environment_variable_name": cls.get_from_env_or_settings(
                "environment_variable_name",
                file_settings,
                cls.environment_variable_name,
            ),
            "default_environment": cls.get_from_env_or_settings(
                "default_environment", file_settings, cls.default_environment
            ),
            "config_file_prefix": cls.get_from_env_or_settings(
                "config_file_prefix", file_settings, cls.config_file_prefix
            ),
            "config_file_suffix": cls.get_from_env_or_settings(
                "config_file_suffix", file_settings, cls.config_file_suffix
            ),
            "environment_prefix": cls.get_from_env_or_settings(
                "environment_prefix", file_settings, cls.environment_prefix
            ),
            "environment_suffix": cls.get_from_env_or_settings(
                "environment_suffix", file_settings, cls.environment_suffix
            ),
            "evaluate_data": cls.get_from_env_or_settings(
                "evaluate_data", file_settings, cls.evaluate_data
            ),
            "config_file_path": cls.get_from_env_or_settings(
                "config_file_path", file_settings, cls.config_file_path
            ),
            "show_logs": cls.get_from_env_or_settings(
                "show_logs", file_settings, cls.show_logs
            ),
            "log_filtered_value": cls.get_from_env_or_settings(
                "log_filtered_value",
                file_settings,
                cls.log_filtered_value,
            ),
            "do_not_read_environment": cls.get_from_env_or_settings(
                "do_not_read_environment",
                file_settings,
                cls.do_not_read_environment,
            ),
            "do_not_read_dotenv": cls.get_from_env_or_settings(
                "do_not_read_dotenv",
                file_settings,
                cls.do_not_read_dotenv,
            ),
            "env_file": cls.get_from_env_or_settings(
                "env_file", file_settings, cls.env_file
            ),
            "load_order": cls.get_from_env_or_settings(
                "load_order", file_settings, DEFAULT_ORDER
            ),
            "env_table": cls.get_from_env_or_settings(
                "env_table", file_settings, cls.env_table
            ),
            "use_environment_layers": cls.get_from_env_or_settings(
                "use_environment_layers", file_settings, cls.use_environment_layers
            ),
            "dotenv_overwrites_memory": cls.get_from_env_or_settings(
                "dotenv_overwrites_memory", file_settings, cls.dotenv_overwrites_memory
            ),
            "warn_if_env_is_missing": cls.get_from_env_or_settings(
                "warn_if_env_is_missing", file_settings, cls.warn_if_env_is_missing
            ),
            "dotenv_encoding": cls.get_from_env_or_settings(
                "dotenv_encoding", file_settings, cls.dotenv_encoding
            ),
        }
        return file_settings, settings



import importlib
import os
from dataclasses import dataclass
from typing import Any, Dict

from loguru import logger

from stela.main import StelaBaseMain
from stela.parsers.dotenv import read_dotenv
from stela.utils import show_value


@dataclass
class StelaDotMain(StelaBaseMain):
    def get_project_settings(self):
        env_data = self.read_env_files()

        module_path = ".".join(self.options.final_loader.split(".")[:-1])
        loader_fn = self.options.final_loader.split(".")[-1]
        module = importlib.import_module(module_path)
        self.settings = getattr(module, loader_fn)(
            options=self.options, env_data=env_data
        )
        logger.debug(
            f"Settings after running loader: "
            f"{[f'{k}={show_value(v, self.options.log_filtered_value)}' for k, v in self.settings.items()]}"
        )

    def read_env_files(self) -> Dict[str, Any]:
        settings = {}
        current_environment = self.options.current_environment
        dotenv_files = [
            self.options.env_file,
            f"{self.options.env_file}.local",
        ]
        if current_environment:
            dotenv_files += [
                f"{self.options.env_file}.{current_environment.lower()}",
                f"{self.options.env_file}.{current_environment.lower()}.local",
            ]
        logger.debug(f"Looking for dotenv files: {dotenv_files}")
        for file in dotenv_files:
            env_settings = read_dotenv(
                config_file_path=self.options.config_file_path,
                env_file=file,
                overwrites_memory=self.options.dotenv_overwrites_memory,
                encoding=self.options.dotenv_encoding,
                verbose=self.options.warn_if_env_is_missing,
                filter_logs=self.options.log_filtered_value,
            )
            for k, v in env_settings.items():
                settings[k] = v
        for k, v in settings.items():
            if k in os.environ and not self.options.dotenv_overwrites_memory:
                continue
            if v is not None:
                os.environ[k] = v
        logger.debug(
            f"Settings after load all dotenv files: "
            f"{[f'{k}={show_value(v, self.options.log_filtered_value)}' for k, v in settings.items()]}"
        )
        return settings



import ast
import os
import re
from copy import deepcopy
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path

import click
from scalpl import Cut


@dataclass
class StelaConverter:
    root_dir: str
    settings_name: str = "settings"
    key_attr_map: dict[str, str] = field(default_factory=dict)
    processed_files: list[str] = field(default_factory=list)
    not_processed_files: list[str] = field(default_factory=list)
    files_found: int = 0
    env_file: str = ".env"
    config_file_path: str = "."

    def run(self, check_only: bool = False):
        for subdir, _, files in os.walk(self.root_dir):
            for file in files:
                filepath = os.path.join(subdir, file)
                if (
                    filepath.endswith(".py")
                    and not filepath.endswith("stela_cmd.py")
                    and not filepath.endswith("stela_converter.py")
                    and not filepath.endswith("stela_init.py")
                    and not filepath.endswith("stela_conf.py")
                ):
                    self.process_python_file(
                        filepath, process_python_file=not check_only
                    )

        self.show_result()

        if self.processed_files:
            self.create_env_file()
            if self.not_processed_files:
                click.secho(
                    "Warning: Some files have old stela data format, but we cannot change "
                    "automatically. Please check.",
                    bold=True,
                    fg="yellow",
                )
            else:
                click.secho(
                    "Success: Done convert files. Please check the .env file.",
                    fg="green",
                    bold=True,
                )
        if not self.files_found:
            click.secho("Success: No old keys found.", fg="green", bold=True)

    def show_result(self):
        # print key-attribute mapping
        if len(self.key_attr_map.keys()) > 0:
            click.secho("Old Keys found and converted names:", fg="white")
            for key, attr in self.key_attr_map.items():
                click.secho(f"* {key} -> {attr}", fg="blue")
        for fn in self.not_processed_files:
            click.secho(
                f"File {fn} has old stela data format, but we cannot change automatically. Please check.",
                bold=True,
                fg="yellow",
            )

    def process_python_file(self, filepath: str, process_python_file: bool = True):
        with open(filepath, "r") as f:
            file_contents = f.read()
            old_contents = deepcopy(file_contents)

        for node in ast.walk(ast.parse(file_contents)):
            if isinstance(node, ast.Import) or isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    if hasattr(node, "module") and "stela" in node.module:
                        self.files_found += 1
                        if process_python_file:
                            self.convert(alias, file_contents, filepath, old_contents)

    def convert(self, alias, file_contents, filepath, old_contents):
        need_backup = False
        self.settings_name = alias.asname if alias.asname else "settings"
        regex = re.compile(
            rf"(?<!\w){self.settings_name}(?!\.env)(?:(?<=\.)[A-Za-z_][A-Za-z0-9_]*)?"
        )
        need_backup = self.find_old_and_new_keys(file_contents, need_backup)
        file_contents = self.replace_in_file_keys(file_contents)
        file_contents = regex.sub(
            lambda match: match.group().replace(self.settings_name, "env"),
            file_contents,
        )
        if need_backup:
            self.backup_python_file(file_contents, filepath, old_contents)
            self.processed_files.append(filepath)
        elif not os.path.exists(f"{filepath}.stela.bak"):
            self.not_processed_files.append(filepath)

    @staticmethod
    def backup_python_file(file_contents, filepath, old_contents):
        backup_file = f"{filepath}.stela.bak"
        if not os.path.exists(backup_file):
            click.secho(f"Creating backup file {backup_file}...", dim=True)
            with open(backup_file, "w") as f:
                f.write(old_contents)
        with open(filepath, "w") as f:
            f.write(file_contents)

    @property
    def settings_dot_regex(self):
        return rf"\b{self.settings_name}\[(['\"])([\w\.]+)\1\]"

    @property
    def settings_get_regex(self):
        return rf"\b{self.settings_name}\.get\((['\"])([\w\.]+)\1\)"

    def replace_in_file_keys(self, file_contents):
        # replace dot with underscore
        file_contents = re.sub(
            self.settings_dot_regex,
            lambda match: f"{self.settings_name}.{match.group(2).replace('.', '_').upper()}",
            file_contents,
            flags=re.ASCII,
        )
        file_contents = re.sub(
            self.settings_get_regex,
            lambda match: f"{self.settings_name}.{match.group(2).replace('.', '_').upper()}",
            file_contents,
            flags=re.ASCII,
        )
        return file_contents

    def find_old_and_new_keys(self, file_contents, need_backup):
        matches = re.findall(
            self.settings_dot_regex,
            file_contents,
            flags=re.ASCII,
        )
        for match in matches:
            need_backup = True
            key = match[1]
            attr = key.replace(".", "_").upper()
            self.key_attr_map[key] = attr
        matches = re.findall(
            self.settings_get_regex,
            file_contents,
            flags=re.ASCII,
        )
        for match in matches:
            need_backup = True
            key = match[1]
            attr = key.replace(".", "_").upper()
            self.key_attr_map[key] = attr
        return need_backup

    def _process_env_data(self, filename: str, origin: Cut):
        can_continue = len(origin.keys()) > 0
        path = os.path.join(Path.cwd(), self.config_file_path)
        file_path = os.path.join(path, filename)
        if can_continue and os.path.exists(file_path):
            can_continue = click.confirm(
                click.style(
                    f"Warning: File {file_path} already exists. Overwrite?", fg="yellow"
                ),
                abort=False,
            )
        if can_continue:
            if not os.path.exists(path):
                os.makedirs(path)
            with open(file_path, "w") as f:
                today = datetime.now().strftime("%Y-%m-%d")
                f.write(
                    f"# Created by Stela in {today} using data from old configuration files.\n"
                )
                for fn in self.processed_files:
                    f.write(f"# {fn}\n")
                f.write("\n")
                all_keys = [k for k in self.key_attr_map]
                all_keys.sort()
                for key in all_keys:
                    line = f"{self.key_attr_map[key]}={origin.get(key, '')}\n"
                    f.write(line)
            click.secho(f"Success: Created file {file_path}.", fg="green")

    def create_env_file(self):
        from stela import settings

        filename = f"{self.env_file}.{settings.stela_current_environment.lower()}.local"
        self._process_env_data(filename, settings)

    def revert(self):
        backup_found = False
        for subdir, _, files in os.walk(self.root_dir):
            for file in files:
                filepath = os.path.join(subdir, file)
                if filepath.endswith(".stela.bak"):
                    backup_found = True
                    click.secho(f"Reverting {filepath}...", dim=True)
                    python_file = filepath.replace(".stela.bak", "")
                    os.remove(python_file)
                    os.rename(filepath, python_file)
        if backup_found:
            click.secho("Success: Done revert convert files.", fg="green", bold=True)



"""Stela Class."""
import os
from ast import literal_eval
from dataclasses import dataclass
from typing import Any, Dict, Optional, Tuple

from _warnings import warn
from loguru import logger
from scalpl import Cut
from scalpl.scalpl import index_error, key_error, split_path, traverse, type_error

from stela.loaders.cut import StelaCutLoader
from stela.main import StelaBaseMain
from stela.parsers.embed import read_embed
from stela.parsers.other_files import read_file
from stela.utils import merge_dicts, show_value


class FakeNone:
    """Fake None for Stela Cut.

    Resolves StelaCut.cut() and StelaCut.pop()
    issues when default=None.
    See: https://github.com/ducdetronquito/scalpl/issues/16
    """


class StelaCut(Cut):  # type: ignore
    """StelaCut subclass."""

    def __init__(self, data: Optional[dict] = None, sep: str = "."):
        super().__init__(data, sep)
        self._options = None
        self._loader = None

    @property
    def stela_options(self) -> "StelaOptions":  # type: ignore
        """Return Stela Options."""
        return self._options

    @stela_options.setter
    def stela_options(self, options: "StelaOptions"):  # type: ignore
        """Save Stela Options in instance."""
        self._options = options

    @property
    def stela_loader(self) -> "StelaLoader":  # type: ignore
        """Return Stela Loader."""
        return self._loader

    @stela_loader.setter
    def stela_loader(self, loader: "StelaLoader"):  # type: ignore
        """Save Stela Loader in instance."""
        self._loader = loader

    @property
    def stela_current_environment(self) -> str:
        """Return Current Environment."""
        return self._options.current_environment

    @property
    def to_dict(self) -> Dict[Any, Any]:
        """Return Cut as Python Dictionary."""
        return dict(self)

    def get(self, path: str, default: Optional[Any] = FakeNone) -> Any:
        """Override Get from Dict code."""
        try:
            return self[path]
        except (KeyError, IndexError) as error:
            if default is not FakeNone:
                return default
            raise error

    def pop(self, path: str, default: Any = FakeNone, *args) -> Any:  # type: ignore
        """Override Pop from Dict code."""
        *keys, last_key = split_path(path, self.sep)

        try:
            item = traverse(data=self.data, keys=keys, original_path=path)
        except (KeyError, IndexError) as error:
            if default is not FakeNone:
                return default
            raise error

        try:
            return item.pop(last_key)
        except (AttributeError, KeyError) as error:
            if default is not FakeNone:
                return default
            raise key_error(last_key, path, error)
        except IndexError as error:
            if default is not FakeNone:
                return default
            raise index_error(last_key, path, error)

    def __getitem__(
        self, path: str, *args: Tuple[Any], **kwargs: Dict[Any, Any]
    ) -> Any:
        """Override getitem from Dict.

        Before invoking  original code, we will
        use the full path received to check for
        a environment variable value, using StelaOptions
        options.

        :param path: full dict path
        :param args: python arguments
        :param kwargs: python keyword arguments
        :return: Any
        """

        warn(
            "Using Stela as Dictionary is deprecated and will be removed in 6.0",
            DeprecationWarning,
            stacklevel=2,
        )

        if "[" not in path:
            environment_variable = self.get_environment_variable_name(path)
            value = self.get_value_from_memory(environment_variable)
            if value:
                if self.stela_options.evaluate_data:
                    try:
                        value = literal_eval(value)
                    except ValueError:
                        from loguru import logger

                        logger.debug(
                            f"Error when evaluating value: {environment_variable}={value[:3]}***"
                        )
                return value

        *keys, last_key = split_path(path, self.sep)
        item = traverse(data=self.data, keys=keys, original_path=path)

        try:
            return item[last_key]
        except KeyError as error:
            raise key_error(last_key, path, error)
        except IndexError as error:
            raise index_error(last_key, path, error)
        except TypeError:
            raise type_error(last_key, path, item)

    def get_value_from_memory(self, environment_variable):
        from loguru import logger

        if (
            self.stela_options.do_not_read_environment
            and environment_variable != self.stela_options.environment_variable_name
        ):
            logger.debug("Ignoring Environment variables in memory.")
            return

        value = os.getenv(environment_variable)
        if value:
            logger.debug(
                f"Using environment value: {environment_variable}="
                f"{show_value(value, self.stela_options.log_filtered_value)}"
            )
        return value

    def get_environment_variable_name(self, path):
        environment_variable = (
            f"{self.stela_options.environment_prefix}"
            f"{path.replace('.', '_')}"
            f"{self.stela_options.environment_suffix}".upper().strip()
        )
        return environment_variable


@dataclass
class StelaCutMain(StelaBaseMain):
    """Stela Cut Main Class.

    Will be removed on 6.0
    """

    def run_embed_loader(self):
        self.loader.embed_data = read_embed(self.options)
        merge_dicts(self.loader.embed_data, self.settings)
        return "pyproject.toml"

    def run_file_loader(self):
        origin, self.loader.file_data = read_file(self.options)
        merge_dicts(self.loader.file_data, self.settings)
        return origin

    def run_custom_loader(self):
        from stela.loaders.cut import StelaCutLoader

        loader = StelaCutLoader()
        if loader.custom_load_function:
            loader.custom_data = loader.custom_load_function(
                data=self.settings, options=self.options
            )
            merge_dicts(loader.custom_data, self.settings)
            return loader.custom_load_function.__name__

    def get_project_settings(self) -> StelaCut:
        """Get project settings running Stela Lifecycle.

        :return: Dict
        """

        self.loader = StelaCutLoader()

        # Pre-Load Phase
        logger.info("Starting Stela Pre-Load Phase...")
        if self.loader.pre_load_function:
            self.loader.pre_data = self.loader.pre_load_function(options=self.options)
            merge_dicts(self.loader.pre_data, self.settings)
            self.log_current_data(origin=self.loader.pre_load_function.__name__)

        # Load Phase
        logger.info(
            f"Starting Stela Load Phase. Order is: [{', '.join(self.options.load_order)}] ..."
        )
        for loader_name in self.options.load_order:
            origin = getattr(self, f"run_{loader_name}_loader")()
            if origin:
                self.log_current_data(origin=origin)

        # Post-Load Phase
        logger.info("Starting Stela Post-Load Phase...")
        if self.loader.post_load_function:
            self.loader.post_data = self.loader.post_load_function(
                data=self.settings, options=self.options
            )
            merge_dicts(self.loader.post_data, self.settings)
            self.log_current_data(origin=self.loader.post_load_function.__name__)

        proxy = StelaCut(self.settings)
        proxy.stela_options = self.options
        proxy.stela_loader = self.loader
        return proxy
